import {inject, singleton} from 'tsyringe'
import {models} from '@tonality/library-common'
import Uuid, {UuidBuffer} from '@tonality/library-uuid';
import {PrismaClient} from '@prisma/client'

export default interface RingtoneService {
	get(id: Uuid): Promise<models.Ringtone>
	browseByComposer(composerUserId: Uuid, skip?: number, take?: number): Promise<models.Ringtone[]>
	browse(skip?: number, take?: number): Promise<models.Ringtone[]>
	search(q?: string): Promise<models.Ringtone[]>
	create(data: Partial<models.Ringtone>): Promise<models.Ringtone>
	update(data: Partial<models.Ringtone>): Promise<models.Ringtone>
	softDelete(id: Uuid): Promise<models.Ringtone>
	undoDelete(id: Uuid): Promise<models.Ringtone>
	hardDelete(id: Uuid): Promise<void>
}

const serializeRingtone = d => {
	return ({
		...d,
		composerUserId: Uuid.from(d.composerUserId).toString(),
		id: Uuid.from(d.id).toString(),
	})
}


@singleton()
export class RingtoneServiceImpl {
	constructor(
		@inject('PrismaClient')
		private readonly prismaClient: PrismaClient,
	) {}

	async browseByComposer(composerUserId: Uuid, skip: number = 0, take: number = 10): Promise<models.Ringtone[]> {
		const rawData = await this.prismaClient.ringtone.findMany({
			where: {
				composer: {
					id: {
						equals: composerUserId,
					}
				}
			},
			skip,
			take,
		})

		return rawData.map(serializeRingtone)
	}

	async get(id: Uuid): Promise<models.Ringtone> {
		const ringtone = await this.prismaClient.ringtone.findUnique({
			where: {
				id,
			}
		})

		if (!ringtone) {
			throw new Error('Ringtone not found!')
		}

		return serializeRingtone(ringtone)
	}

	async browse(skip: number = 0, take: number = 10): Promise<models.Ringtone[]> {
		const rawData = await this.prismaClient.ringtone.findMany({
			skip,
			take,
		})

		return rawData.map(serializeRingtone)
	}

	async search(q: string = ''): Promise<models.Ringtone[]> {
		return this.prismaClient.ringtone.findMany({
			where: {
				name: {
					contains: q,
				},
			},
		})
	}

	async create(data: Partial<models.Ringtone>): Promise<models.Ringtone> {
		const { createdAt, updatedAt, deletedAt, composerUserId, name, ...safeData } = data
		const rawData = await this.prismaClient.ringtone.create({
			data: {
				...safeData,
				id: Uuid.new().toBuffer(),
				composerUserId: (composerUserId as UuidBuffer).toBuffer(),
				name: name as string,
			},
		})
		return serializeRingtone(rawData)
	}

	async update(data: Partial<models.Ringtone>): Promise<models.Ringtone> {
		const { createdAt, updatedAt, deletedAt, ...safeData } = data
		const rawData = await this.prismaClient.ringtone.update({
			where: {
				id: (data.id as UuidBuffer).toBuffer(),
			},
			data: {
				...safeData,
				updatedAt: new Date(),
			},
		})
		return serializeRingtone(rawData)
	}

	async softDelete(id: Uuid): Promise<models.Ringtone> {
		const rawData = await this.prismaClient.ringtone.update({
			where: {
				id,
			},
			data: {
				deletedAt: new Date(),
			},
		})
		return serializeRingtone(rawData)
	}

	async undoDelete(id: Uuid): Promise<models.Ringtone> {
		const rawData = this.prismaClient.ringtone.update({
			where: {
				id,
			},
			data: {
				deletedAt: null,
			},
		})
		return serializeRingtone(rawData)
	}

	async hardDelete(id: Uuid): Promise<void> {
		await this.prismaClient.ringtone.delete({
			where: {
				id,
			},
		})
	}
}
