import {inject, singleton} from 'tsyringe'
import {PrismaClient} from '@prisma/client'
import { models } from '@tonality/library-common'
import Uuid from '@tonality/library-uuid'
import PasswordService from '../password/service'

export default interface UserService {
	get(id: Uuid): Promise<models.UserProfile>
	getByUsername(username: string): Promise<models.User>
	getProfileByUsername(username: string): Promise<models.UserProfile>
	browse(skip?: number, take?: number): Promise<models.UserProfile[]>
	search(q?: string): Promise<models.UserProfile[]>
	create(profile: Partial<models.UserProfile>, username: string, newPassword: string, confirmNewPassword: string): Promise<models.UserProfile>
	updateProfile(profile: Partial<models.UserProfile>): Promise<models.UserProfile>
	updatePassword(id: Uuid, oldPassword: string, newPassword: string, confirmNewPassword: string): Promise<void>
	delete(id: Uuid): Promise<void>
}

@singleton()
export class UserServiceImpl implements UserService {
	constructor(
		@inject('PrismaClient')
		private readonly prismaClient: PrismaClient,
		@inject('PasswordService')
		private readonly passwordService: PasswordService
	) {}

	async get(id: Uuid): Promise<models.UserProfile> {
		const user = await this.prismaClient.userProfile.findUnique({
			where: {
				userId: id,
			},
		})

		if (!user) {
			throw new Error('User not found!')
		}

		return user
	}

	async getByUsername(username: string): Promise<models.User> {
		const user = await this.prismaClient.user.findUnique({
			where: {
				username,
			},
		})

		if (!user) {
			throw new Error('User not found!')
		}

		return user
	}

	async getProfileByUsername(username: string): Promise<models.UserProfile> {
		const user = await this.prismaClient.userProfile.findFirst({
			where: {
				user: {
					username,
				},
			},
		})

		if (!user) {
			throw new Error('User not found!')
		}

		return user
	}

	async browse(skip: number = 0, take: number = 10): Promise<models.UserProfile[]> {
		return this.prismaClient.userProfile.findMany({
			skip,
			take,
		})
	}

	async search(q: string = ''): Promise<models.UserProfile[]> {
		return this.prismaClient.userProfile.findMany({
			where: {
				user: {
					username: {
						contains: q,
					},
				},
			},
		})
	}

	async create(profile: Partial<models.UserProfile>, username: string, newPassword: string, confirmNewPassword: string): Promise<models.UserProfile> {
		if (newPassword !== confirmNewPassword) {
			throw new Error('Passwords do not match!')
		}

		const newUserId = Uuid.new()
		const password = await this.passwordService.hash(newPassword)

		await this.prismaClient.user.create({
			data: {
				id: newUserId,
				username,
				password,
			},
		})

		return this.prismaClient.userProfile.create({
			data: {
				...profile,
				userId: newUserId,
			},
		})
	}

	async updateProfile(profile: Partial<models.UserProfile>): Promise<models.UserProfile> {
		const { userId, ...etcProfile } = profile
		return this.prismaClient.userProfile.update({
			where: {
				userId: userId as Uuid,
			},
			data: etcProfile,
		})
	}

	async updatePassword(id: Uuid, oldPassword: string, newPassword: string, confirmNewPassword: string): Promise<void> {
		const user = await this.prismaClient.user.findUnique({
			where: {
				id,
			},
		})

		if (!user) {
			throw new Error('Invalid request!')
		}

		const oldPasswordsMatch = await this.passwordService.compare(user.password, oldPassword)
		if (!oldPasswordsMatch) {
			throw new Error('Invalid request!')
		}

		if (newPassword !== confirmNewPassword) {
			throw new Error('Passwords do not match!')
		}
		const password = await this.passwordService.hash(newPassword)
		await this.prismaClient.user.update({
			where: {
				id,
			},
			data: {
				password,
			},
		})
	}

	async delete(id: Uuid): Promise<void> {
		await this.prismaClient.user.delete({
			where: {
				id,
			},
		})
	}
}
